@@ -139,6 +139,7 @@ update_default_source (GvcApplet *applet)
         if (stream != NULL) {
                 gvc_stream_status_icon_set_mixer_stream (applet->priv->input_status_icon,
                                                          stream);
+                maybe_show_status_icons(applet);
         } else {
                 g_warning ("Unable to get default source");
         }
@@ -153,6 +154,7 @@ update_default_sink (GvcApplet *applet)
         if (stream != NULL) {
                 gvc_stream_status_icon_set_mixer_stream (applet->priv->output_status_icon,
                                                          stream);
+                maybe_show_status_icons(applet);
         } else {
                 g_warning ("Unable to get default sink");
         }
@@ -217,7 +219,7 @@ gvc_applet_constructor (GType                  type,
 
         self = GVC_APPLET (object);
 
-        self->priv->control = gvc_mixer_control_new ();
+        self->priv->control = gvc_mixer_control_new ("GNOME Volume Control Applet");
         g_signal_connect (self->priv->control,
                           "ready",
                           G_CALLBACK (on_control_ready),
