@@ -32,12 +32,14 @@
 #include "gvc-balance-bar.h"
 
 #define SCALE_SIZE 128
+#define ADJUSTMENT_MAX_NORMAL 65536.0 /* PA_VOLUME_NORM */
 
 #define GVC_BALANCE_BAR_GET_PRIVATE(o) (G_TYPE_INSTANCE_GET_PRIVATE ((o), GVC_TYPE_BALANCE_BAR, GvcBalanceBarPrivate))
 
 struct GvcBalanceBarPrivate
 {
         GvcChannelMap *channel_map;
+        GvcBalanceType btype;
         GtkWidget     *scale_box;
         GtkWidget     *start_box;
         GtkWidget     *end_box;
@@ -52,7 +54,8 @@ struct GvcBalanceBarPrivate
 enum
 {
         PROP_0,
-        PROP_CHANNEL_MAP
+        PROP_CHANNEL_MAP,
+        PROP_BALANCE_TYPE,
 };
 
 static void     gvc_balance_bar_class_init (GvcBalanceBarClass *klass);
@@ -68,6 +71,8 @@ static gboolean on_scale_button_release_event (GtkWidget      *widget,
 static gboolean on_scale_scroll_event         (GtkWidget      *widget,
                                                GdkEventScroll *event,
                                                GvcBalanceBar  *bar);
+static void on_adjustment_value_changed       (GtkAdjustment *adjustment,
+                                               GvcBalanceBar *bar);
 
 G_DEFINE_TYPE (GvcBalanceBar, gvc_balance_bar, GTK_TYPE_HBOX)
 
@@ -79,7 +84,7 @@ _scale_box_new (GvcBalanceBar *bar)
         GtkWidget            *sbox;
         GtkWidget            *ebox;
         GtkAdjustment        *adjustment = bar->priv->adjustment;
-        char                 *str;
+        char                 *str_lower, *str_upper;
 
         bar->priv->scale_box = box = gtk_hbox_new (FALSE, 6);
         priv->scale = gtk_hscale_new (priv->adjustment);
@@ -98,18 +103,35 @@ _scale_box_new (GvcBalanceBar *bar)
 
         gtk_box_pack_start (GTK_BOX (box), priv->scale, TRUE, TRUE, 0);
 
-        str = g_strdup_printf ("<small>%s</small>", C_("balance", "Left"));
-        gtk_scale_add_mark (GTK_SCALE (priv->scale), adjustment->lower , 
-                            GTK_POS_BOTTOM, str);
-        g_free (str);
+        switch (bar->priv->btype) {
+        case BALANCE_TYPE_RL:
+                str_lower = g_strdup_printf ("<small>%s</small>", C_("balance", "Left"));
+                str_upper = g_strdup_printf ("<small>%s</small>", C_("balance", "Right"));
+                break;
+        case BALANCE_TYPE_FR:
+                str_lower = g_strdup_printf ("<small>%s</small>", C_("balance", "Rear"));
+                str_upper = g_strdup_printf ("<small>%s</small>", C_("balance", "Front"));
+                break;
+        case BALANCE_TYPE_LFE:
+                str_lower = g_strdup_printf ("<small>%s</small>", C_("balance", "Minimum"));
+                str_upper = g_strdup_printf ("<small>%s</small>", C_("balance", "Maximum"));
+                break;
+        default:
+                g_assert_not_reached ();
+        }
 
-        str = g_strdup_printf ("<small>%s</small>", C_("balance", "Right"));
-        gtk_scale_add_mark (GTK_SCALE (priv->scale),  adjustment->upper, 
-                            GTK_POS_BOTTOM, str);
-        g_free (str);
+        gtk_scale_add_mark (GTK_SCALE (priv->scale), adjustment->lower,
+                            GTK_POS_BOTTOM, str_lower);
+        g_free (str_lower);
+        gtk_scale_add_mark (GTK_SCALE (priv->scale),  adjustment->upper,
+                            GTK_POS_BOTTOM, str_upper);
+        g_free (str_upper);
 
-        gtk_scale_add_mark (GTK_SCALE (priv->scale), (adjustment->upper - adjustment->lower)/2 + adjustment->lower, 
-                            GTK_POS_BOTTOM, NULL);
+        if (bar->priv->btype != BALANCE_TYPE_LFE) {
+                gtk_scale_add_mark (GTK_SCALE (priv->scale),
+                                    (adjustment->upper - adjustment->lower)/2 + adjustment->lower,
+                                    GTK_POS_BOTTOM, NULL);
+        }
 
         bar->priv->end_box = ebox = gtk_hbox_new (FALSE, 6);
         gtk_box_pack_start (GTK_BOX (box), ebox, FALSE, FALSE, 0);
@@ -160,6 +182,58 @@ gvc_balance_bar_set_size_group (GvcBalanceBar *bar,
         gtk_widget_queue_draw (GTK_WIDGET (bar));
 }
 
+static const char *
+btype_to_string (guint btype)
+{
+        switch (btype) {
+        case BALANCE_TYPE_RL:
+                return "Balance";
+        case BALANCE_TYPE_FR:
+                return "Fade";
+                break;
+        case BALANCE_TYPE_LFE:
+                return "LFE";
+        default:
+                g_assert_not_reached ();
+        }
+        return NULL;
+}
+
+static void
+update_level_from_map (GvcBalanceBar *bar,
+                       GvcChannelMap *map)
+{
+        const gdouble *volumes;
+        gdouble val;
+
+        g_debug ("Volume changed (for %s bar)", btype_to_string (bar->priv->btype));
+
+        volumes = gvc_channel_map_get_volume (map);
+        switch (bar->priv->btype) {
+        case BALANCE_TYPE_RL:
+                val = volumes[BALANCE];
+                break;
+        case BALANCE_TYPE_FR:
+                val = volumes[FADE];
+                break;
+        case BALANCE_TYPE_LFE:
+                val = volumes[LFE];
+                break;
+        default:
+                g_assert_not_reached ();
+        }
+
+        gtk_adjustment_set_value (bar->priv->adjustment, val);
+}
+
+static void
+on_channel_map_volume_changed (GvcChannelMap  *map,
+                               gboolean        set,
+                               GvcBalanceBar  *bar)
+{
+        update_level_from_map (bar, map);
+}
+
 static void
 gvc_balance_bar_set_channel_map (GvcBalanceBar *bar,
                                  GvcChannelMap *map)
@@ -167,13 +241,84 @@ gvc_balance_bar_set_channel_map (GvcBalanceBar *bar,
         g_return_if_fail (GVC_BALANCE_BAR (bar));
 
         if (bar->priv->channel_map != NULL) {
+                g_signal_handlers_disconnect_by_func (G_OBJECT (bar->priv->channel_map),
+                                                      on_channel_map_volume_changed, bar);
                 g_object_unref (bar->priv->channel_map);
         }
         bar->priv->channel_map = g_object_ref (map);
 
+        update_level_from_map (bar, map);
+
+        g_signal_connect (G_OBJECT (map), "volume-changed",
+                          G_CALLBACK (on_channel_map_volume_changed), bar);
+
         g_object_notify (G_OBJECT (bar), "channel-map");
 }
 
+static void
+gvc_balance_bar_set_balance_type (GvcBalanceBar *bar,
+                                  GvcBalanceType btype)
+{
+        GtkWidget *frame;
+
+        g_return_if_fail (GVC_BALANCE_BAR (bar));
+
+        bar->priv->btype = btype;
+        if (bar->priv->btype != BALANCE_TYPE_LFE) {
+                bar->priv->adjustment = GTK_ADJUSTMENT (gtk_adjustment_new (0.0,
+                                                                            -1.0,
+                                                                            1.0,
+                                                                            0.5,
+                                                                            0.5,
+                                                                            0.0));
+        } else {
+                bar->priv->adjustment = GTK_ADJUSTMENT (gtk_adjustment_new (0.0,
+                                                                            0.0,
+                                                                            ADJUSTMENT_MAX_NORMAL,
+                                                                            ADJUSTMENT_MAX_NORMAL/100.0,
+                                                                            ADJUSTMENT_MAX_NORMAL/10.0,
+                                                                            0.0));
+        }
+
+        g_object_ref_sink (bar->priv->adjustment);
+        g_signal_connect (bar->priv->adjustment,
+                          "value-changed",
+                          G_CALLBACK (on_adjustment_value_changed),
+                          bar);
+
+        switch (btype) {
+        case BALANCE_TYPE_RL:
+                bar->priv->label = gtk_label_new_with_mnemonic (_("_Balance:"));
+                break;
+        case BALANCE_TYPE_FR:
+                bar->priv->label = gtk_label_new_with_mnemonic (_("_Fade:"));
+                break;
+        case BALANCE_TYPE_LFE:
+                bar->priv->label = gtk_label_new_with_mnemonic (_("_Subwoofer:"));
+                break;
+        default:
+                g_assert_not_reached ();
+        }
+        gtk_misc_set_alignment (GTK_MISC (bar->priv->label),
+                                0.0,
+                                0.0);
+        /* frame */
+        frame = gtk_frame_new (NULL);
+        gtk_frame_set_shadow_type (GTK_FRAME (frame), GTK_SHADOW_NONE);
+        gtk_container_add (GTK_CONTAINER (bar), frame);
+
+        /* box with scale */
+        bar->priv->scale_box = _scale_box_new (bar);
+        gtk_container_add (GTK_CONTAINER (frame), bar->priv->scale_box);
+        gtk_widget_show_all (frame);
+
+        gtk_widget_set_direction (bar->priv->scale, GTK_TEXT_DIR_LTR);
+        gtk_label_set_mnemonic_widget (GTK_LABEL (bar->priv->label),
+                                       bar->priv->scale);
+
+        g_object_notify (G_OBJECT (bar), "balance-type");
+}
+
 static void
 gvc_balance_bar_set_property (GObject       *object,
                               guint          prop_id,
@@ -186,6 +331,9 @@ gvc_balance_bar_set_property (GObject       *object,
         case PROP_CHANNEL_MAP:
                 gvc_balance_bar_set_channel_map (self, g_value_get_object (value));
                 break;
+        case PROP_BALANCE_TYPE:
+                gvc_balance_bar_set_balance_type (self, g_value_get_int (value));
+                break;
         default:
                 G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
                 break;
@@ -215,14 +363,7 @@ gvc_balance_bar_constructor (GType                  type,
                              guint                  n_construct_properties,
                              GObjectConstructParam *construct_params)
 {
-        GObject       *object;
-        GvcBalanceBar *self;
-
-        object = G_OBJECT_CLASS (gvc_balance_bar_parent_class)->constructor (type, n_construct_properties, construct_params);
-
-        self = GVC_BALANCE_BAR (object);
-
-        return object;
+        return G_OBJECT_CLASS (gvc_balance_bar_parent_class)->constructor (type, n_construct_properties, construct_params);
 }
 
 static void
@@ -242,35 +383,18 @@ gvc_balance_bar_class_init (GvcBalanceBarClass *klass)
                                                               "The channel map",
                                                               GVC_TYPE_CHANNEL_MAP,
                                                               G_PARAM_READWRITE));
+        g_object_class_install_property (object_class,
+                                         PROP_BALANCE_TYPE,
+                                         g_param_spec_int ("balance-type",
+                                                           "balance type",
+                                                           "Whether the balance is right-left or front-rear",
+                                                           BALANCE_TYPE_RL, NUM_BALANCE_TYPES - 1, BALANCE_TYPE_RL,
+                                                           G_PARAM_READWRITE|G_PARAM_CONSTRUCT_ONLY));
 
         g_type_class_add_private (klass, sizeof (GvcBalanceBarPrivate));
 }
 
 
-static gboolean
-on_left (pa_channel_position_t p)
-{
-    return
-        p == PA_CHANNEL_POSITION_FRONT_LEFT ||
-        p == PA_CHANNEL_POSITION_REAR_LEFT ||
-        p == PA_CHANNEL_POSITION_FRONT_LEFT_OF_CENTER ||
-        p == PA_CHANNEL_POSITION_SIDE_LEFT ||
-        p == PA_CHANNEL_POSITION_TOP_FRONT_LEFT ||
-        p == PA_CHANNEL_POSITION_TOP_REAR_LEFT;
-}
-
-static gboolean
-on_right (pa_channel_position_t p)
-{
-    return
-        p == PA_CHANNEL_POSITION_FRONT_RIGHT ||
-        p == PA_CHANNEL_POSITION_REAR_RIGHT ||
-        p == PA_CHANNEL_POSITION_FRONT_RIGHT_OF_CENTER ||
-        p == PA_CHANNEL_POSITION_SIDE_RIGHT ||
-        p == PA_CHANNEL_POSITION_TOP_FRONT_RIGHT ||
-        p == PA_CHANNEL_POSITION_TOP_REAR_RIGHT;
-}
-
 static gboolean
 on_scale_button_press_event (GtkWidget      *widget,
                              GdkEventButton *event,
@@ -300,132 +424,122 @@ on_scale_scroll_event (GtkWidget      *widget,
 
         value = gtk_adjustment_get_value (bar->priv->adjustment);
 
-        if (event->direction == GDK_SCROLL_UP) {
-                if (value + 0.01 > 1.0)
-                        value = 1.0;
-                else
-                        value = value + 0.01;
-                gtk_adjustment_set_value (bar->priv->adjustment, value);
-        } else if (event->direction == GDK_SCROLL_DOWN) {
-                if (value - 0.01 < 0)
-                        value = 0.0;
-                else
-                        value = value - 0.01;
-                gtk_adjustment_set_value (bar->priv->adjustment, value);
+        if (bar->priv->btype == BALANCE_TYPE_LFE) {
+                if (event->direction == GDK_SCROLL_UP) {
+                        if (value + ADJUSTMENT_MAX_NORMAL/100.0 > ADJUSTMENT_MAX_NORMAL)
+                                value = ADJUSTMENT_MAX_NORMAL;
+                        else
+                                value = value + ADJUSTMENT_MAX_NORMAL/100.0;
+                } else if (event->direction == GDK_SCROLL_DOWN) {
+                        if (value - ADJUSTMENT_MAX_NORMAL/100.0 < 0)
+                                value = 0.0;
+                        else
+                                value = value - ADJUSTMENT_MAX_NORMAL/100.0;
+                }
+        } else {
+                if (event->direction == GDK_SCROLL_UP) {
+                        if (value + 0.01 > 1.0)
+                                value = 1.0;
+                        else
+                                value = value + 0.01;
+                } else if (event->direction == GDK_SCROLL_DOWN) {
+                        if (value - 0.01 < 0)
+                                value = 0.0;
+                        else
+                                value = value - 0.01;
+                }
         }
+        gtk_adjustment_set_value (bar->priv->adjustment, value);
 
         return TRUE;
 }
 
+/* FIXME remove when we depend on a newer PA */
+static pa_cvolume *
+gvc_pa_cvolume_set_position (pa_cvolume *cv, const pa_channel_map *map, pa_channel_position_t t, pa_volume_t v) {
+        unsigned c;
+        gboolean good = FALSE;
+
+        g_assert(cv);
+        g_assert(map);
+
+        g_return_val_if_fail(pa_cvolume_compatible_with_channel_map(cv, map), NULL);
+        g_return_val_if_fail(t < PA_CHANNEL_POSITION_MAX, NULL);
+
+        for (c = 0; c < map->channels; c++)
+                if (map->map[c] == t) {
+                        cv->values[c] = v;
+                        good = TRUE;
+                }
+
+        return good ? cv : NULL;
+}
+
 static void
 on_adjustment_value_changed (GtkAdjustment *adjustment,
                              GvcBalanceBar *bar)
 {
         gdouble                val;
-        gdouble               *gains;
-        pa_channel_position_t *positions;
-        guint                  num_channels;
-        guint                  i;
-        gdouble                left_v;
-        gdouble                center_v;
-        gdouble                right_v;
+        pa_cvolume             cv;
+        const pa_channel_map  *pa_map;
 
+        if (bar->priv->channel_map == NULL)
+                return;
+
+        cv = *gvc_channel_map_get_cvolume (bar->priv->channel_map);
         val = gtk_adjustment_get_value (adjustment);
 
-        if (bar->priv->channel_map == NULL) {
-                return;
+        pa_map = gvc_channel_map_get_pa_channel_map (bar->priv->channel_map);
+
+        switch (bar->priv->btype) {
+        case BALANCE_TYPE_RL:
+                pa_cvolume_set_balance (&cv, pa_map, val);
+                break;
+        case BALANCE_TYPE_FR:
+                pa_cvolume_set_fade (&cv, pa_map, val);
+                break;
+        case BALANCE_TYPE_LFE:
+                gvc_pa_cvolume_set_position (&cv, pa_map, PA_CHANNEL_POSITION_LFE, val);
+                break;
         }
 
-        left_v = 1.0;
-        right_v = 1.0;
-
-        if (val > 0) {
-                left_v = 1.0 - val;
-        } else if (val < 0) {
-                right_v = 1.0 - ABS(val);
-        }
-        center_v = (left_v + right_v) / 2.0;
-
-        num_channels = gvc_channel_map_get_num_channels (bar->priv->channel_map);
-        positions = gvc_channel_map_get_positions (bar->priv->channel_map);
-        gains = gvc_channel_map_get_gains (bar->priv->channel_map);
-
-        for (i = 0; i < num_channels; i++) {
-                if (on_left (positions[i])) {
-                        gains[i] = left_v;
-                } else if (on_right (positions[i])) {
-                        gains[i] = right_v;
-                } else {
-                        gains[i] = center_v;
-                }
-        }
-
-        gvc_channel_map_gains_changed (bar->priv->channel_map);
+        gvc_channel_map_volume_changed (bar->priv->channel_map, &cv, TRUE);
 }
 
 static void
 gvc_balance_bar_init (GvcBalanceBar *bar)
 {
-        GtkWidget *frame;
-
         bar->priv = GVC_BALANCE_BAR_GET_PRIVATE (bar);
-
-        bar->priv->adjustment = GTK_ADJUSTMENT (gtk_adjustment_new (0.0,
-                                                                    -1.0,
-                                                                    1.0,
-                                                                    0.05,
-                                                                    0.1,
-                                                                    0.1));
-        g_object_ref_sink (bar->priv->adjustment);
-        g_signal_connect (bar->priv->adjustment,
-                          "value-changed",
-                          G_CALLBACK (on_adjustment_value_changed),
-                          bar);
-
-        bar->priv->label = gtk_label_new_with_mnemonic (_("_Balance:"));
-        gtk_misc_set_alignment (GTK_MISC (bar->priv->label),
-                                0.0,
-                                0.5);
-        /* frame */
-        frame = gtk_frame_new (NULL);
-        gtk_frame_set_shadow_type (GTK_FRAME (frame), GTK_SHADOW_NONE);
-        gtk_container_add (GTK_CONTAINER (bar), frame);
-
-        /* box with scale */
-        bar->priv->scale_box = _scale_box_new (bar);
-        gtk_container_add (GTK_CONTAINER (frame), bar->priv->scale_box);
-        gtk_widget_show_all (frame);
-
-        gtk_widget_set_direction (bar->priv->scale, GTK_TEXT_DIR_LTR);
-        gtk_label_set_mnemonic_widget (GTK_LABEL (bar->priv->label),
-                                       bar->priv->scale);
 }
 
 static void
 gvc_balance_bar_finalize (GObject *object)
 {
-        GvcBalanceBar *balance_bar;
+        GvcBalanceBar *bar;
 
         g_return_if_fail (object != NULL);
         g_return_if_fail (GVC_IS_BALANCE_BAR (object));
 
-        balance_bar = GVC_BALANCE_BAR (object);
+        bar = GVC_BALANCE_BAR (object);
 
-        g_return_if_fail (balance_bar->priv != NULL);
+        g_return_if_fail (bar->priv != NULL);
 
-        if (balance_bar->priv->channel_map != NULL) {
-                g_object_unref (balance_bar->priv->channel_map);
+        if (bar->priv->channel_map != NULL) {
+                g_signal_handlers_disconnect_by_func (G_OBJECT (bar->priv->channel_map),
+                                                      on_channel_map_volume_changed, bar);
+                g_object_unref (bar->priv->channel_map);
         }
 
         G_OBJECT_CLASS (gvc_balance_bar_parent_class)->finalize (object);
 }
 
 GtkWidget *
-gvc_balance_bar_new (GvcChannelMap *channel_map)
+gvc_balance_bar_new (GvcChannelMap *channel_map, GvcBalanceType btype)
 {
         GObject *bar;
         bar = g_object_new (GVC_TYPE_BALANCE_BAR,
                             "channel-map", channel_map,
+                            "balance-type", btype,
                             NULL);
         return GTK_WIDGET (bar);
 }
